@@ -29,6 +29,7 @@ char sandbox_lib[SB_PATH_MAX];
 
 typedef struct {
 	bool show_access_violation, on, active, testing, verbose, debug;
+	sandbox_method_t method;
 	char *ld_library_path;
 	char **prefixes[5];
 	int num_prefixes[5];
@@ -94,6 +95,7 @@ void libsb_init(void)
 	sbcontext.verbose = is_env_on(ENV_SANDBOX_VERBOSE);
 	sbcontext.debug = is_env_on(ENV_SANDBOX_DEBUG);
 	sbcontext.testing = is_env_on(ENV_SANDBOX_TESTING);
+	sbcontext.method = get_sandbox_method();
 	if (sbcontext.testing) {
 		const char *ldpath = getenv("LD_LIBRARY_PATH");
 		if (ldpath)
@@ -101,6 +103,11 @@ void libsb_init(void)
 	}
 }
 
+sandbox_method_t get_sandbox_method(void)
+{
+	return parse_sandbox_method(getenv(ENV_SANDBOX_METHOD));
+}
+
 /* resolve_dirfd_path - get the path relative to a dirfd
  *
  * return value:
@@ -125,14 +132,15 @@ int resolve_dirfd_path(int dirfd, const char *path, char *resolved_path,
 	save_errno();
 
 	size_t at_len = resolved_path_len - 1 - 1 - (path ? strlen(path) : 0);
-	if (trace_pid)
-	    sprintf(resolved_path, "/proc/%i/fd/%i", trace_pid, dirfd);
-	else
-	    /* If /proc was mounted by a process in a different pid namespace,
-	     * getpid cannot be used to create a valid /proc/<pid> path. Instead
-	     * use sb_get_fd_dir() which works in any case.
-	     */
-	    sprintf(resolved_path, "%s/%i", sb_get_fd_dir(), dirfd);
+	if (trace_pid) {
+		sprintf(resolved_path, "/proc/%i/fd/%i", trace_pid, dirfd);
+	} else {
+		/* If /proc was mounted by a process in a different pid namespace,
+		 * getpid cannot be used to create a valid /proc/<pid> path. Instead
+		 * use sb_get_fd_dir() which works in any case.
+		 */
+		sprintf(resolved_path, "%s/%i", sb_get_fd_dir(), dirfd);
+	}
 	ssize_t ret = readlink(resolved_path, resolved_path, at_len);
 	if (ret == -1) {
 		/* see comments at end of check_syscall() */
@@ -663,33 +671,31 @@ static int check_prefixes(char **prefixes, int num_prefixes, const char *path)
 }
 
 /* Is this a func that works on symlinks, and is the file a symlink ? */
-static bool symlink_func(int sb_nr, int flags, const char *abs_path)
+static bool symlink_func(int sb_nr, int flags)
 {
-	struct stat st;
-
 	/* These funcs always operate on symlinks */
-	if (!(sb_nr == SB_NR_UNLINK   ||
-	      sb_nr == SB_NR_UNLINKAT ||
-	      sb_nr == SB_NR_LCHOWN   ||
-	      sb_nr == SB_NR_REMOVE   ||
-	      sb_nr == SB_NR_RENAME   ||
-	      sb_nr == SB_NR_RENAMEAT ||
-	      sb_nr == SB_NR_RMDIR    ||
-	      sb_nr == SB_NR_SYMLINK  ||
-	      sb_nr == SB_NR_SYMLINKAT))
-	{
-		/* These funcs sometimes operate on symlinks */
-		if (!((sb_nr == SB_NR_FCHOWNAT ||
-		       sb_nr == SB_NR_FCHMODAT ||
-		       sb_nr == SB_NR_UTIMENSAT) &&
-		      (flags & AT_SYMLINK_NOFOLLOW)))
-			return false;
-	}
-
-	if (-1 != lstat(abs_path, &st) && S_ISLNK(st.st_mode))
+	if (sb_nr == SB_NR_UNLINK       ||
+	    sb_nr == SB_NR_UNLINKAT     ||
+	    sb_nr == SB_NR_LCHOWN       ||
+	    sb_nr == SB_NR_LREMOVEXATTR ||
+	    sb_nr == SB_NR_LSETXATTR    ||
+	    sb_nr == SB_NR_REMOVE       ||
+	    sb_nr == SB_NR_RENAME       ||
+	    sb_nr == SB_NR_RENAMEAT     ||
+	    sb_nr == SB_NR_RENAMEAT2    ||
+	    sb_nr == SB_NR_RMDIR        ||
+	    sb_nr == SB_NR_SYMLINK      ||
+	    sb_nr == SB_NR_SYMLINKAT)
 		return true;
-	else
-		return false;
+
+	/* These funcs sometimes operate on symlinks */
+	if ((sb_nr == SB_NR_FCHOWNAT ||
+	     sb_nr == SB_NR_FCHMODAT ||
+	     sb_nr == SB_NR_UTIMENSAT) &&
+	    (flags & AT_SYMLINK_NOFOLLOW))
+		return true;
+
+	return false;
 }
 
 static int check_access(sbcontext_t *sbcontext, int sb_nr, const char *func,
@@ -698,7 +704,7 @@ static int check_access(sbcontext_t *sbcontext, int sb_nr, const char *func,
 	int old_errno = errno;
 	int result = 0;
 	int retval;
-	bool sym_func = symlink_func(sb_nr, flags, abs_path);
+	bool sym_func = symlink_func(sb_nr, flags);
 
 	retval = check_prefixes(sbcontext->deny_prefixes,
 		sbcontext->num_deny_prefixes, abs_path);
@@ -767,6 +773,8 @@ static int check_access(sbcontext_t *sbcontext, int sb_nr, const char *func,
 	    sb_nr == SB_NR_LCHOWN      ||
 	    sb_nr == SB_NR_LINK        ||
 	    sb_nr == SB_NR_LINKAT      ||
+	    sb_nr == SB_NR_LREMOVEXATTR||
+	    sb_nr == SB_NR_LSETXATTR   ||
 	    sb_nr == SB_NR_LUTIMES     ||
 	    sb_nr == SB_NR_MKDIR       ||
 	    sb_nr == SB_NR_MKDIRAT     ||
@@ -785,9 +793,12 @@ static int check_access(sbcontext_t *sbcontext, int sb_nr, const char *func,
 	    sb_nr == SB_NR_MKSTEMPS64  ||
 	    sb_nr == SB_NR_OPEN_WR     ||
 	    sb_nr == SB_NR_REMOVE      ||
+	    sb_nr == SB_NR_REMOVEXATTR ||
 	    sb_nr == SB_NR_RENAME      ||
 	    sb_nr == SB_NR_RENAMEAT    ||
+	    sb_nr == SB_NR_RENAMEAT2   ||
 	    sb_nr == SB_NR_RMDIR       ||
+	    sb_nr == SB_NR_SETXATTR    ||
 	    sb_nr == SB_NR_SYMLINK     ||
 	    sb_nr == SB_NR_SYMLINKAT   ||
 	    sb_nr == SB_NR_TRUNCATE    ||
@@ -852,20 +863,6 @@ static int check_access(sbcontext_t *sbcontext, int sb_nr, const char *func,
 			goto out;
 		}
 
-		/* A very common bug (apparently) is for .py[co] files to fall out
-		 * of sync with their .py source files.  Rather than trigger a hard
-		 * failure, let's just whine about it.  Once python itself gets
-		 * sorted out, we can drop this #256953.
-		 */
-		size_t len = strlen(resolv_path);
-		if (len > 4) {
-			const char *py = resolv_path + len - 4;
-			if (!strcmp(py, ".pyc") || !strcmp(py, ".pyo")) {
-				sbcontext->show_access_violation = false;
-				goto out;
-			}
-		}
-
 		/* If we are here, and still no joy, and its the access() call,
 		 * do not log it, but just return -1 */
 		if (sb_nr == SB_NR_ACCESS_WR) {
@@ -902,7 +899,7 @@ static int check_syscall(sbcontext_t *sbcontext, int sb_nr, const char *func,
 	 * itself does not dereference.  This speeds things up and avoids updating
 	 * the atime implicitly. #415475
 	 */
-	if (symlink_func(sb_nr, flags, absolute_path))
+	if (symlink_func(sb_nr, flags))
 		resolved_path = absolute_path;
 	else
 		resolved_path = resolve_path(file, 1);
@@ -1122,7 +1119,7 @@ bool before_syscall_open_char(int dirfd, int sb_nr, const char *func, const char
 typedef struct {
 	const char *name;
 	size_t len;
-	char *value;
+	const char *value;
 } env_pair;
 #define ENV_PAIR(x, n, v) [x] = { .name = n, .len = sizeof(n) - 1, .value = v, }
 
@@ -1170,6 +1167,7 @@ struct sb_envp_ctx sb_new_envp(char **envp, bool insert)
 		ENV_PAIR(11, ENV_SANDBOX_DEBUG, NULL),
 		ENV_PAIR(12, "LD_LIBRARY_PATH", NULL),
 		ENV_PAIR(13, ENV_SANDBOX_TESTING, NULL),
+		ENV_PAIR(14, ENV_SANDBOX_METHOD, NULL),
 	};
 	size_t num_vars = ARRAY_SIZE(vars);
 	char *found_vars[num_vars];
@@ -1242,6 +1240,8 @@ struct sb_envp_ctx sb_new_envp(char **envp, bool insert)
 		vars[12].value = sbcontext.ld_library_path;
 		vars[13].value = "1";
 	}
+	if (sbcontext.method != SANDBOX_METHOD_ANY)
+		vars[14].value = str_sandbox_method(sbcontext.method);
 
 	char ** my_env = NULL;
 	if (!insert) {
