@@ -178,20 +178,6 @@ static char *do_peekstr(unsigned long lptr)
 	}
 }
 
-static const char *strcld_chld(int cld)
-{
-	switch (cld) {
-#define C(c) case CLD_##c: return "CLD_"#c;
-	C(CONTINUED)
-	C(DUMPED)
-	C(EXITED)
-	C(KILLED)
-	C(TRAPPED)
-	C(STOPPED)
-#undef C
-	default: return "CLD_???";
-	}
-}
 /* strsignal() translates the string when i want C define */
 static const char *strsig(int sig)
 {
@@ -214,47 +200,6 @@ static const char *strsig(int sig)
 	}
 }
 
-static void trace_child_signal(int signo, siginfo_t *info, void *context)
-{
-	sb_debug("got sig %s(%i): code:%s(%i) status:%s(%i)",
-		strsig(signo), signo,
-		strcld_chld(info->si_code), info->si_code,
-		strsig(info->si_status), info->si_status);
-
-	switch (info->si_code) {
-		case CLD_DUMPED:
-		case CLD_KILLED:
-			trace_exit(128 + info->si_status);
-
-		case CLD_EXITED:
-			__sb_debug(" = %i\n", info->si_status);
-			trace_exit(info->si_status);
-
-		case CLD_TRAPPED:
-			switch (info->si_status) {
-				case SIGSTOP:
-					kill(trace_pid, SIGCONT);
-				case SIGTRAP:
-				case SIGCONT:
-					return;
-			}
-
-			/* For whatever signal the child caught, let's ignore it and
-			 * continue on.  If it aborted, segfaulted, whatever, that's
-			 * its problem, not ours, so don't whine about it.  We just
-			 * have to be sure to bubble it back up.  #265072
-			 */
-			do_ptrace(PTRACE_CONT, NULL, (void *)(long)info->si_status);
-			return;
-	}
-
-	sb_eerror("ISE:trace_child_signal: child (%i) signal %s(%i), code %s(%i), status %s(%i)\n",
-		trace_pid,
-		strsig(signo), signo,
-		strcld_chld(info->si_code), info->si_code,
-		strsig(info->si_status), info->si_status);
-}
-
 static const struct syscall_entry *lookup_syscall_in_tbl(const struct syscall_entry *tbl, int nr)
 {
 	while (tbl->name)
@@ -272,6 +217,7 @@ struct syscall_state {
 	bool (*pre_check)(const char *func, const char *pathname, int dirfd);
 };
 
+/* Check syscall that only takes a path as its |ibase| argument. */
 static bool _trace_check_syscall_C(struct syscall_state *state, int ibase)
 {
 	char *path = do_peekstr(trace_arg(state->regs, ibase));
@@ -288,6 +234,7 @@ static bool _trace_check_syscall_C(struct syscall_state *state, int ibase)
 	free(path);
 	return ret;
 }
+/* Check syscall that only takes a path as its first argument. */
 static bool trace_check_syscall_C(struct syscall_state *state)
 {
 	return _trace_check_syscall_C(state, 1);
@@ -310,20 +257,24 @@ static bool __trace_check_syscall_DCF(struct syscall_state *state, int ibase, in
 	free(path);
 	return ret;
 }
-static bool _trace_check_syscall_DCF(struct syscall_state *state, int ibase)
+/* Check syscall that takes a dirfd & path starting at |ibase| argument, and flags at |fbase|. */
+static bool _trace_check_syscall_DCF(struct syscall_state *state, int ibase, int fbase)
 {
-	int flags = trace_arg(state->regs, ibase + 2);
+	int flags = trace_arg(state->regs, fbase);
 	return __trace_check_syscall_DCF(state, ibase, flags);
 }
+/* Check syscall that takes a dirfd, path, and flags as its first 3 arguments. */
 static bool trace_check_syscall_DCF(struct syscall_state *state)
 {
-	return _trace_check_syscall_DCF(state, 1);
+	return _trace_check_syscall_DCF(state, 1, 3);
 }
 
+/* Check syscall that takes a dirfd & path starting at |ibase| argument. */
 static bool _trace_check_syscall_DC(struct syscall_state *state, int ibase)
 {
 	return __trace_check_syscall_DCF(state, ibase, 0);
 }
+/* Check syscall that takes a dirfd & path as its first 2 arguments (but no flags). */
 static bool trace_check_syscall_DC(struct syscall_state *state)
 {
 	return _trace_check_syscall_DC(state, 1);
@@ -354,7 +305,7 @@ static bool trace_check_syscall(const struct syscall_entry *se, void *regs)
 	state.regs = regs;
 	state.nr = nr = se->sys;
 	state.func = name = se->name;
-	if (nr == SB_NR_UNDEF)          goto done;
+	if (!SB_NR_IS_DEFINED(se->nr))  goto done;
 	else if (nr == SB_NR_MKDIR)     state.pre_check = sb_mkdirat_pre_check;
 	else if (nr == SB_NR_MKDIRAT)   state.pre_check = sb_mkdirat_pre_check;
 	else if (nr == SB_NR_UNLINK)    state.pre_check = sb_unlinkat_pre_check;
@@ -362,22 +313,24 @@ static bool trace_check_syscall(const struct syscall_entry *se, void *regs)
 	else                            state.pre_check = NULL;
 
 	/* Hmm, add these functions to the syscall table and avoid this if() ? */
-	if (nr == SB_NR_UNDEF)          goto done;
-	else if (nr == SB_NR_CHMOD)     return  trace_check_syscall_C  (&state);
+	     if (nr == SB_NR_CHMOD)     return  trace_check_syscall_C  (&state);
 	else if (nr == SB_NR_CHOWN)     return  trace_check_syscall_C  (&state);
 	else if (nr == SB_NR_CREAT)     return  trace_check_syscall_C  (&state);
-	else if (nr == SB_NR_FCHMODAT)  return  trace_check_syscall_DCF(&state);
-	else if (nr == SB_NR_FCHOWNAT)  return  trace_check_syscall_DCF(&state);
+	/* NB: Linux syscall does not have a flags argument. */
+	else if (nr == SB_NR_FCHMODAT)  return  trace_check_syscall_DC (&state);
+	else if (nr == SB_NR_FCHOWNAT)  return _trace_check_syscall_DCF(&state, 1, 5);
 	else if (nr == SB_NR_FUTIMESAT) return  trace_check_syscall_DC (&state);
 	else if (nr == SB_NR_LCHOWN)    return  trace_check_syscall_C  (&state);
 	else if (nr == SB_NR_LINK)      return _trace_check_syscall_C  (&state, 2);
-	else if (nr == SB_NR_LINKAT)    return _trace_check_syscall_DCF(&state, 3);
+	else if (nr == SB_NR_LINKAT)    return _trace_check_syscall_DCF(&state, 3, 5);
 	else if (nr == SB_NR_MKDIR)     return  trace_check_syscall_C  (&state);
 	else if (nr == SB_NR_MKDIRAT)   return  trace_check_syscall_DC (&state);
 	else if (nr == SB_NR_MKNOD)     return  trace_check_syscall_C  (&state);
 	else if (nr == SB_NR_MKNODAT)   return  trace_check_syscall_DC (&state);
 	else if (nr == SB_NR_RENAME)    return  trace_check_syscall_C  (&state) &&
 	                                       _trace_check_syscall_C  (&state, 2);
+	else if (nr == SB_NR_RENAMEAT2) return  trace_check_syscall_DC (&state) &&
+	                                       _trace_check_syscall_DC (&state, 3);
 	else if (nr == SB_NR_RENAMEAT)  return  trace_check_syscall_DC (&state) &&
 	                                       _trace_check_syscall_DC (&state, 3);
 	else if (nr == SB_NR_RMDIR)     return  trace_check_syscall_C  (&state);
@@ -389,7 +342,14 @@ static bool trace_check_syscall(const struct syscall_entry *se, void *regs)
 	else if (nr == SB_NR_UNLINKAT)  return  trace_check_syscall_DCF(&state);
 	else if (nr == SB_NR_UTIME)     return  trace_check_syscall_C  (&state);
 	else if (nr == SB_NR_UTIMES)    return  trace_check_syscall_C  (&state);
-	else if (nr == SB_NR_UTIMENSAT) return _trace_check_syscall_DCF(&state, 1);
+
+	else if (nr == SB_NR_UTIMENSAT) {
+ utimensat:
+		return _trace_check_syscall_DCF(&state, 1, 4);
+	} else if (nr == SB_NR___UTIMENSAT64) {
+		state.nr = SB_NR_UTIMENSAT;
+		goto utimensat;
+	}
 
 	else if (nr == SB_NR_ACCESS) {
 		char *path = do_peekstr(trace_arg(regs, 1));
@@ -441,36 +401,88 @@ static void trace_loop(void)
 {
 	trace_regs regs;
 	bool before_exec, before_syscall, fake_syscall_ret;
+	unsigned event;
 	long ret;
-	int nr, status;
-	const struct syscall_entry *se, *tbl_after_fork;
+	int status, sig;
+	const struct syscall_entry *tbl_after_fork;
+	void *data;
 
 	before_exec = true;
 	before_syscall = false;
 	fake_syscall_ret = false;
 	tbl_after_fork = NULL;
+	data = NULL;
 	do {
-		ret = do_ptrace(PTRACE_SYSCALL, NULL, NULL);
+		ret = do_ptrace(PTRACE_SYSCALL, NULL, data);
+		data = NULL;
 		waitpid(trace_pid, &status, 0);
 
-		if (before_exec) {
-			unsigned event = ((unsigned)status >> 16);
-			if (event == PTRACE_EVENT_EXEC) {
-				_sb_debug("hit exec!");
-				before_exec = false;
-			} else
+		event = (unsigned)status >> 16;
+
+		if (WIFSIGNALED(status)) {
+			int sig = WTERMSIG(status);
+			sb_debug("signaled %s %i", strsig(sig), sig);
+			kill(getpid(), sig);
+			trace_exit(128 + sig);
+
+		} else if (WIFEXITED(status)) {
+			ret = WEXITSTATUS(status);
+			sb_debug("exited %li", ret);
+			trace_exit(ret);
+
+		}
+
+		sb_assert(WIFSTOPPED(status));
+		sig = WSTOPSIG(status);
+
+		switch (event) {
+		case 0:
+			if (sig != (SIGTRAP | 0x80)) {
+				/* For whatever signal the child caught, let's ignore it and
+				 * continue on.  If it aborted, segfaulted, whatever, that's
+				 * its problem, not ours, so don't whine about it.  We just
+				 * have to be sure to bubble it back up.  #265072
+				 *
+				 * The next run of this loop should see the WIFSIGNALED status
+				 * and we'll exit then.
+				 */
+				sb_debug("passing signal through %s (%i)", strsig(sig), sig);
+				data = (void *)(uintptr_t)(sig);
+				continue;
+			}
+
+			if (before_exec) {
 				_sb_debug("waiting for exec; status: %#x", status);
+				continue;
+			}
+			break;
+
+		case PTRACE_EVENT_EXEC:
+			__sb_debug("hit exec!");
+			before_exec = false;
 			ret = trace_get_regs(&regs);
 			tbl_after_fork = trace_check_personality(&regs);
 			continue;
+
+		case PTRACE_EVENT_EXIT:
+			/* We'll tell the process to resume, which should make it exit,
+			 * and then we'll pick up its exit status and exit above.
+			 */
+			__sb_debug(" exit event!\n");
+			continue;
+
+		default:
+			sb_ebort("ISE: unhandle ptrace signal %s (%i) event %u\n",
+			         strsig(sig), sig, event);
 		}
 
 		ret = trace_get_regs(&regs);
-		nr = trace_get_sysnum(&regs);
 
-		se = lookup_syscall_in_tbl(tbl_after_fork, nr);
-		ret = trace_get_regs(&regs);
 		if (before_syscall) {
+			/* NB: The kernel guarantees syscall NR is valid only on entry. */
+			int nr = trace_get_sysnum(&regs);
+			const struct syscall_entry *se = lookup_syscall_in_tbl(tbl_after_fork, nr);
+
 			_sb_debug("%s:%i", se ? se->name : "IDK", nr);
 			if (!trace_check_syscall(se, &regs)) {
 				sb_debug_dyn("trace_loop: forcing EPERM after %s\n", se->name);
@@ -498,19 +510,14 @@ static void trace_loop(void)
 	} while (1);
 }
 
-void trace_main(const char *filename, char *const argv[])
+void trace_main(void)
 {
-	struct sigaction sa, old_sa;
-
-	sa.sa_flags = SA_RESTART | SA_SIGINFO;
-	sa.sa_sigaction = trace_child_signal;
-	sigaction(SIGCHLD, &sa, &old_sa);
-
-	sb_debug_dyn("trace_main: tracing: %s\n", filename);
+	struct sigaction old_sa, sa = { .sa_handler = SIG_DFL, };
 
 	if (trace_pid)
 		sb_ebort("ISE: trace code assumes multiple threads are not forking\n");
 
+	sigaction(SIGCHLD, &sa, &old_sa);
 	trace_pid = fork();
 	if (unlikely(trace_pid == -1)) {
 		sb_ebort("ISE: vfork() failed: %s\n", strerror(errno));
@@ -518,7 +525,12 @@ void trace_main(const char *filename, char *const argv[])
 		sb_debug("parent waiting for child (pid=%i) to signal", trace_pid);
 		waitpid(trace_pid, NULL, 0);
 		do_ptrace(PTRACE_SETOPTIONS, NULL,
-			(void *)(PTRACE_O_TRACESYSGOOD | PTRACE_O_TRACEEXEC));
+			(void *)(uintptr_t)(
+				PTRACE_O_EXITKILL |
+				PTRACE_O_TRACEEXEC |
+				PTRACE_O_TRACEEXIT |
+				PTRACE_O_TRACESYSGOOD
+			));
 		sb_close_all_fds();
 		trace_loop();
 		sb_ebort("ISE: child should have quit, as should we\n");
@@ -536,7 +548,7 @@ void trace_main(const char *filename, char *const argv[])
 #undef _trace_possible
 #define _trace_possible(data) false
 
-void trace_main(const char *filename, char *const argv[])
+void trace_main(void)
 {
 	/* trace_possible() triggers a warning for us */
 }
@@ -572,8 +584,14 @@ static char *flatten_args(char *const argv[])
 
 bool trace_possible(const char *filename, char *const argv[], const void *data)
 {
-	if (_trace_possible(data))
-		return true;
+	if (_trace_possible(data)) {
+		/* If we're in an environment like QEMU where ptrace doesn't work, then
+		 * don't try to use it.  If ptrace does work, this should fail with ESRCH.
+		 */
+		errno = 0;
+		ptrace(PTRACE_CONT, 0, NULL, NULL);
+		return errno == ENOSYS ? false : true;
+	}
 
 	char *args = flatten_args(argv);
 	sb_eqawarn("Unable to trace static ELF: %s: %s\n", filename, args);
